#!/usr/bin/perl
use strict;
use warnings;
use diagnostics;
use Curses;
use Curses::UI;
use File::Find::Rule;
use Getopt::Long;
use Fcntl ':flock';
use Data::Dumper;
use vars qw( $cui $screen $background $config_file $config $new_screen $domain_list $selected_domain );
use subs qw( menu init init_widgets bind_keys read_config write_config about exit_dialog
  dialog_modify_remove_choice remove_domain modify_domain load_file save_as );

sub menu {
  init();
  init_widgets();
  bind_keys();
  $screen->focus();
  $cui->mainloop;
}

menu();
1;

#################################################################################
### INITIALIZE
sub init {
  ### init
  GetOptions ('config=s' => \$config_file ) or die $!;
  ($config_file) = File::Find::Rule->file()->name('Config.pm')->in( map { $_ . '/PerlWebmail' } @INC ) unless $config_file;
  die "Cannot find PerlWebmail/Config.pm, try passing via --config" unless $config_file;
  read_config();
  open STDERR,">/dev/null";
  #open STDERR,">err";
}

sub init_widgets {
  $cui = Curses::UI->new(
    -clear_on_exit => 1,
  ) unless $cui;
  $cui->add(
    'menu','Menubar',
    -menu => [
          {-label => 'File', -submenu => [ { -label => 'Open ^O', -value => \&load_file },
          { -label => 'Save ^S', -value => \&write_config },
          { -label => 'Save As ^W', -value => \&save_as },
          { -label => 'Quit ^Q', -value => \&exit_dialog },
      ]},
      {-label => 'Domain', -submenu => [
          { -label => 'Show Domains ^E', -value => sub{ $screen->focus();} },
          { -label => 'New ^N', -value => sub{ $new_screen->focus();} },
      ]},
      {-label => 'Help', -submenu => [
          { -label => 'About', -value => \&about },
      ]},
    ],
  );
  $background = $cui->add(
    'background','Window',
    -border => 1, -y => -1, -height => 3,
  );
  $background->add(
    'help_menu','Label',
    -text => ' ' x 80,
  );

  $screen = $cui->add(
    'screen','Window',
    -padtop => 1,
    -border => 1,
    -ipad   => 0,
    -padbottom  => 3,
  );
  $domain_list = $screen->add(
    'domain_list', 'Listbox',
    #-y          => 1,
    -ipad       => 1,
    -padbottom  => 2,
    -values     => [keys %{$config}],
    -labels     => { map { $_ => $_ } keys %$config },
    -width      => 60,
    -height     => 20,
    -border     => 1,
    -title      => 'Domains',
    -vscrollbar => 1,
    -onchange   => sub { my $widget = shift; $selected_domain = $widget->get(); dialog_modify_remove_choice($selected_domain,$widget) if $selected_domain; },
    -onfocus    => sub { $background->getobj('help_menu')->text('CTRL+X: Menu  CTRL+Q: Quit'), },
  );

  $new_screen = $cui->add(
    'new_screen','Window',
    -padtop => 1,
    -border => 1,
    -ipad   => 0,
    -padbottom  => 3,
    -onfocus => sub { $background->getobj('help_menu')->text('Add a new domain to the configuration'); },
  );
  $new_screen->add(
    undef,'Label',
    -text => 'Enter the fully qualified domain name.  It will inherit settings from [default].',
  );
  my $ndm_tb = $new_screen->add(
    'new_domain_name', 'TextEntry',
    -singleline => 1,
    -sbborder => 5,
    -y => 2,
    -width => 40,
  );
  $new_screen->add(
    undef, 'Buttonbox',
    -y => 3,
    -buttons => [
      { -label => '< create >', -value => 'create',
        -onpress => sub {
          my $ndn = $ndm_tb->get();
          $config->{$ndn} = { %{$config->{default}} };
          $domain_list->title('Domains (*)');
          $domain_list->values( [keys %{$config}] );
          $screen->focus();
        }, 
      },
      { -label => '< cancel >', -value => 'cancel', -onpress => sub { $domain_list->values( [keys %{$config}] ); $screen->focus(); }, },
    ],
  );
  
}

sub bind_keys {
  $cui->set_binding(\&exit_dialog,"\cQ","\cC");
  $cui->set_binding(sub{ $cui->getobj('menu')->focus},"\cX",KEY_F(10));
  $cui->set_binding(sub{ $new_screen->focus(); },"\cN");
  $cui->set_binding(sub{ $screen->focus(); },"\cE");
  $cui->set_binding( \&write_config,"\cS");
  $cui->set_binding( \&load_file,"\cO");
  $cui->set_binding( \&save_as,"\cW");
}

#################################################################################
### SUPPORT
sub read_config {
  unless( -w $config_file ){
    #warn "Cannot wrote to this file: $config_file";
    $cui = Curses::UI->new(
      -clear_on_exit => 1,
    );
    my $answer = $cui->error(
      -message => "Could not load configuration:\n[$config_file]\n\nDid you try passing in the file\nlocation with the --config flag?\n\nContinue anyway?",
      -buttons => ['yes', 'no'],
      -title   => 'Error',
    );
    exit(0) unless $answer;
  }else{
    require $config_file;
    $config = $PerlWebmail::Config::config if $PerlWebmail::Config::config;
  }
}

sub write_config {
                                                                                                                                                                
  open(FH,">$config_file") or die $!;
  flock(FH,LOCK_EX);
                                                                                                                                                                
  print FH "package PerlWebmail::Config;\n\n";
  print FH Data::Dumper->Dump([$config],[qw($config)]);
  print FH qq{\n1;\n};
                                                                                                                                                                
  flock(FH,LOCK_UN);
  close(FH);

  $domain_list->title('Domains');

}

#################################################################################
### WIDGETS
sub about {
  $cui->dialog(
    -title => 'About PerlWebmail::Admin',
    -message => "Perl Webmail Administration\n"
      . "by Jason Woodward <woodward\@jaos.org>\n"
      . "\n"
      . "This is the curses configuration\n"
      . "editor for Perl Webmail.\n"
      . "http://software.jaos.org"
  );
}

sub exit_dialog {
  my $return = $cui->dialog(
    -title    => 'Confirm',
    -buttons  => ['yes','no'],
    -message  => 'Quit?',
    -onfocus => sub { $background->getobj('help_menu')->text('Are you sure you want to quit?'); },
  );
  if( $return ){
    exit(0);
  }
}

sub dialog_modify_remove_choice {
  my $domain = shift;
  my $widget = shift;
  my $rv = $widget->root->dialog(
    -message => "Would you like to modify or remove this domain?",
    -buttons => [
      { -label => '< modify >', -value => 'modify', },
      { -label => '< remove >', -value => 'remove', },
      { -label => '< cancel >', -value => 'cancel', },
    ],
    -title   => "$domain",
    -onfocus => sub { $background->getobj('help_menu')->text("Modify or Remove $domain ?"); },
  );
  remove_domain($domain) if $rv eq 'remove';
  modify_domain($domain) if $rv eq 'modify';
}

sub remove_domain {
  my $domain = shift;
  if($domain eq 'default'){
    $cui->error(
      -title => 'Error',
      -message => 'Cannot remove default',
    );
    return;
  }
  my $return = $cui->dialog(
    -title    => 'Confirm',
    -buttons  => ['yes','no'],
    -message  => "Remove domain: [$domain]?",
  );
  if( $return ){
    delete $config->{$domain};
    $domain_list->title('Domains (*)');
    $domain_list->values( [keys %{$config}] ),
  }
}

sub modify_domain {
  my $domain = shift;
  my $modify_window = $cui->add(
    undef,'Window',
    -title => "Modify $domain",
    -padtop => 1,
    -border => 1,
    -ipad   => 0,
    -padbottom  => 3,
  );

  ### storage structure for new settings
  my %new_settings;

  $modify_window->add( 'logout_url_label','Label', -y => 0, -width => 20, -border => 0, -text => 'Logout URL',);
  $new_settings{logouturl} = $modify_window->add(
    'logout_url','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 0, -x => $modify_window->getobj('logout_url_label')->width + 2,
    -text => $config->{$domain}->{logouturl}, -regexp => '/^http\w?|ftp:\/\/[\w\.\-]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('Logout URL'); },
  );

  $modify_window->add( 'cssurl_label','Label', -y => 1, -width => 20, -border => 0, -text => 'CSS URL');
  $new_settings{cssurl} = $modify_window->add(
    'cssurl','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 1, -x => $modify_window->getobj('cssurl_label')->width + 2,
    -text => $config->{$domain}->{cssurl}, -regexp => '/^http\w?|ftp:\/\/[\w\.\-]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('CSS URL'); },
  );

  $modify_window->add( 'imagelocation_label', 'Label', -y => 2, -width => 20, -border => 0, -text => 'Image URL' );
  $new_settings{imagelocation} = $modify_window->add(
    'imagelocation','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 2, -x => $modify_window->getobj('imagelocation_label')->width + 2,
    -text => $config->{$domain}->{imagelocation}, -regexp => '/^http\w?|ftp:\/\/[\w\.\-]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('Image URL'); },
  );

  $modify_window->add( 'popauthmode_label','Label', -y => 3, -width => 20, -border => 0, -text => 'POP3 Auth Mode',);
  my %auth_methods = ( APOP => 0, AUTO => 1, LOGIN => 2, );
  $new_settings{popauthmode} = $modify_window->add(
    'popauthmode','Popupmenu',
    -y => 3, -x => $modify_window->getobj('popauthmode_label')->width + 2,
    -values => [ sort keys %auth_methods ],
    -labels => { map +{ $_, $_ }, keys %auth_methods },
    -selected => $auth_methods{ $config->{$domain}->{popauthmode} },
    -onfocus => sub { $background->getobj('help_menu')->text('Pop authentication mode'); },
  );

  $modify_window->add( 'popserver_label','Label', -y => 4, -width => 20, -border => 0, -text => 'POP3 Server',);
  $new_settings{popserver} = $modify_window->add(
    'popserver','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 4, -x => $modify_window->getobj('popserver_label')->width + 2,
    -text => $config->{$domain}->{popserver}, -regexp => '/^[a-zA-Z0-9\-\.]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('POP3 Server name'); },
  );

  $modify_window->add( 'popserverport_label','Label', -y => 5, -width => 20, -border => 0, -text => 'POP3 Port',);
  $new_settings{popserverport} = $modify_window->add(
    'popserverport','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 5, -x => $modify_window->getobj('popserverport_label')->width + 2,
    -text => $config->{$domain}->{popserverport}, -regexp => '/^\d+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('POP3 Server port'); },
  );


  $modify_window->add( 'smtpserver_label','Label', -y => 6, -width => 20, -border => 0, -text => 'SMTP Server',);
  $new_settings{smtpserver} = $modify_window->add(
    'smtpserver','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 6, -x => $modify_window->getobj('smtpserver_label')->width + 2,
    -text => $config->{$domain}->{smtpserver}, -regexp => '/^[a-zA-Z0-9\-\.]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('SMTP Server name'); },
  );

  $modify_window->add( 'smtpserverport_label','Label', -y => 7, -width => 20, -border => 0, -text => 'SMTP Port',);
  $new_settings{smtpserverport} = $modify_window->add(
    'smtpserverport','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 7, -x => $modify_window->getobj('smtpserverport_label')->width + 2,
    -text => $config->{$domain}->{smtpserverport}, -regexp => '/^\d+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('SMTP Server port'); },
  );


  $modify_window->add( 'hostname_label','Label', -y => 8, -width => 20, -border => 0, -text => 'Hostname',);
  $new_settings{hostname} = $modify_window->add(
    'hostname','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 8, -x => $modify_window->getobj('hostname_label')->width + 2,
    -text => $config->{$domain}->{hostname}, -regexp => '/^[a-zA-Z0-9\-\.]+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('Base hostname'); },
  );

  $modify_window->add( 'user_storage_label','Label', -y => 9, -width => 20, -border => 0, -text => 'User Storage',);
  $new_settings{user_storage} = $modify_window->add(
    'user_storage','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 9, -x => $modify_window->getobj('user_storage_label')->width + 2,
    -text => $config->{$domain}->{user_storage},
    -onfocus => sub { $background->getobj('help_menu')->text('Directory for user storage'); },
  );

  $modify_window->add( 'charset_label','Label', -y => 10, -width => 20, -border => 0, -text => 'Charset',);
  $new_settings{charset} = $modify_window->add(
    'charset','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 10, -x => $modify_window->getobj('charset_label')->width + 2,
    -text => $config->{$domain}->{charset},
    -onfocus => sub { $background->getobj('help_menu')->text('Default charset'); },
  );


  $modify_window->add( 'maxmesslistnum_label','Label', -y => 11, -width => 20, -border => 0, -text => 'Max msg in list',);
  $new_settings{maxmesslistnum} = $modify_window->add(
    'maxmesslistnum','TextEntry',
    -border => 0, -maxlength => 60, -width => 60,
    -y => 11, -x => $modify_window->getobj('maxmesslistnum_label')->width + 2,
    -text => $config->{$domain}->{maxmesslistnum}, -regexp => '/^\d+$/',
    -onfocus => sub { $background->getobj('help_menu')->text('Maximun number of msgs to show at once'); },
  );

  $modify_window->add( 'defaultsort_label','Label', -y => 12, -width => 20, -border => 0, -text => 'Default Sort',);
  my %sorts = (date => 0,daterev => 1,from => 2,fromrev => 3,size => 4,sizerev => 5,subject => 6,subjectrev => 7,);
  $new_settings{defaultsort} = $modify_window->add(
    'defaultsort','Popupmenu',
    -y => 12, -x => $modify_window->getobj('defaultsort_label')->width + 2,
    -values => [ sort keys %sorts ],
    -labels => { map +{ $_, $_ }, keys %sorts },
    -selected => $sorts{ $config->{$domain}->{defaultsort} },
    -onfocus => sub { $background->getobj('help_menu')->text('Default sort method'); },
  );

  $modify_window->add( 'encryptlinksandforms_label','Label', -y => 13, -width => 20, -border => 0, -text => 'Encrypt Links/Forms',);
  $new_settings{encryptlinksandforms} = $modify_window->add(
    'encryptlinksandforms','Popupmenu',
    -y => 13, -x => $modify_window->getobj('encryptlinksandforms_label')->width + 2,
    -values => [ qw( 0 1 ) ],
    -labels => { '0' => 'No', '1' => 'Yes' },
    -selected => $config->{$domain}->{encryptlinksandforms} == 1 ? 1 : 0,
    -onfocus => sub { $background->getobj('help_menu')->text('Encrypt links and forms'); },
  );

  $modify_window->add( 'debug_label','Label', -y => 14, -width => 20, -border => 0, -text => 'Debug',);
  $new_settings{debug} = $modify_window->add(
    'debug','Popupmenu',
    -y => 14, -x => $modify_window->getobj('debug_label')->width + 2,
    -values => [ qw( 0 1 ) ],
    -labels => { '0' => 'No', '1' => 'Yes' },
    -selected => $config->{$domain}->{debug} == 1 ? 1 : 0,
    -onfocus => sub { $background->getobj('help_menu')->text('Enable debugging'); },
  );

  $modify_window->add(
    undef, 'Buttonbox',
    -y => 15,
    -buttons => [
      { -label => '< modify >', -value => 'modify',
        -onpress => sub {
          foreach my $setting ( keys %new_settings ){
            $config->{$domain}->{$setting} = $new_settings{$setting}->get();
          }
          $domain_list->title('Domains (*)');
          $domain_list->values( [keys %{$config}] );
          $screen->focus();
        }, 
      },
      { -label => '< cancel >', -value => 'cancel', -onpress => sub { $domain_list->values( [keys %{$config}] ); $screen->focus(); }, },
    ],
    -onfocus => sub { $background->getobj('help_menu')->text('Commit or cancel changes'); },
  );

  $modify_window->focus();
}

sub load_file {
  my $new_config_file = $cui->loadfilebrowser(
    -title => "Load Config File",
    -mask  => [
      ['.',      'All files (*)'      ],
      ['\.pl$', 'Perl files (*.pl)' ],
      ['\.pm$',  'Perl modules (*.pm)'],
    ],
    -onfocus => sub { $background->getobj('help_menu')->text('Select file to load'); },
  );
  if( $new_config_file ){
    $config_file = $new_config_file;
    read_config();
    $domain_list->values( [keys %{$config}] ),
    $domain_list->focus();
  }
}

sub save_as {
  $config_file = $cui->savefilebrowser(
    -title => 'Save As',
    -file => $config_file,
    -onfocus => sub { $background->getobj('help_menu')->text('Save file as'); },
  );
  write_config();
}

__END__

=head1 NAME

PerlWebmail::ConfigAdmin - Curses based configuration tool

=head1 SYNOPSYS

 # with PerlWebmail::ConfigAdmin somewhere within @INC
 perldoc PerlWebmail::ConfigAdmin
 # or
 perldoc ./PerlWebmail/ConfigAdmin.pm

 # To start,
 perl -MPerlWebmail::ConfigAdmin -e menu
 # or 
 perl ./PerlWebmail/ConfigAdmin.pm

=head1 Consideration

 PerlWebmail::ConfigAdmin will accept the location for the Config.pm from
 the command line via the --config option, or it will try to auto detect
 the location of the file using File::Find::Rule.  If the specified or
 autodetected file are not readable, PerlWebmail::ConfigAdmin will
 complain and prompt to continue on anyway.

 Any changes made to the PerlWebmail::Config can be saved within the 
 PerlWebmail::ConfigAdmin menu by using the menu bar or CTRL+S.

 The default domain must be left, as any subsequently created domains
 inherit their settings from it.  Also, the default domain is a fall
 back mechanism in case the fully qualified name requested is not present
 within PerlWebmail::Config.

=head1 Arguments

=over 4

=item I<--config> (Optional) Use this configuration file.  Tries to autodetect otherwise

=back

=head1 AUTHOR

 Jason Woodward <woodwardj@jaos.org>

=cut
