package PerlWebmail::UserConfig::File;
use base PerlWebmail::UserConfig;
use vars qw( $VERSION $NAME);
use strict;
use warnings;
require Data::Dumper;
require Safe;
$NAME = 'PerlWebmail::UserConfig::File';
$VERSION = '3.0.7';

sub init {
  my $self = shift;
  return 0 unless ref($self);
  my (%args) = @_;

  $self->{debug} = $args{-debug} if $args{-debug};
  warn ref($self) . " init: called\n" if $self->{debug};

  die "No user storage directory or email specified" unless ( $args{-user_storage_dir} && $args{-email} );
  die "Base User storage dir does not exist: $args{-user_storage_dir}" unless (
    ( -d $args{-user_storage_dir} )
    && ( -w $args{-user_storage_dir} )
    );

  $self->{user_storage_dir} = $args{-user_storage_dir};
  $self->{email} = $args{-email};

  chdir $self->{user_storage_dir};
  if( -d $args{-email} ){
    die "$args{-email} does not existing within base storage" unless ( -w $args{-email} );
    chdir $args{-email};
    unless ( -f 'Config.pm'){
      $self->write_config();
    }else{
      $self->read_config();
    }
  }else{
    mkdir $args{-email};
    chdir $args{-email};
    $self->write_config();
  }
  return 1;
}

sub read_config {
  my $self = shift;
  return unless ref($self);
  warn ref($self) . " read_config: called\n" if $self->{debug};

  chdir $self->{user_storage_dir};
  chdir $self->{email};

  ### using eval require
  ### (pollutes PerlWebmail::UserConfig::Config under mod_perl)
  #eval " require './Config.pm'" if ( -f './Config.pm');
  #warn ref($self) . " read_config: error [$@]\n" if ($self->{debug} && $@);
  #$self->{config} = $PerlWebmail::UserConfig::Config::config unless ( $@ || !$PerlWebmail::UserConfig::Config::config);

  ### using Safe
  ### (keeps PerlWebmail::UserConfig::Config clean under mod_perl)
  open(CONF,'<','./Config.pm') or die "Failed to open configuration";
  ### the grep clears out some legacy formatting
  my $data = join('', grep { !/\A(?:package|1)/} <CONF>);
  close(CONF);
  $self->{config} = Safe->new()->reval($data);

}

sub write_config {
  my $self = shift;
  return unless ref($self);
  warn ref($self) . " write_config: called\n" if $self->{debug};

  chdir $self->{user_storage_dir};
  chdir $self->{email};

  open(CONFIG,">Config.pm") or die ref($self) . " write_config: failed to write config [$!] [$@]\n";

  print CONFIG qq|
### this is the user configuration for Jaos PerlWebmail
#this_is_legacy package PerlWebmail::UserConfig::Config;
|;

  my $config = $self->{config};
  print CONFIG Data::Dumper->Dump([$config],[qw($config)]);
  #print CONFIG qq{\n1;\n}; ### legacy
  close(CONFIG);

}

1;

__END__

=head1 NAME

PerlWebmail::UserConfig::File - User Configuration object with File based data store

=head1 SYNOPSIS

 ### you can call explicitly or via PerlWebmail::UserConfig
 use PerlWebmail::UserConfig;
 # or
 use PerlWebmail::UserConfig::File;

 my $config_obj = new PerlWebmail::UserConfig::File(
  -user_storage_dir => '/var/lib/perlwebmail',
  -email => 'johndoe@publicdomain.com',
 );

 ### retrieve config 
 $hashref_config = $config_obj->config();
 ### set config using external datasource
 $config_obj->config( $hashref_config );
 ### set config option
 $config_obj->set( key => value );

 ### add and remove folders
 $config_obj->addfolder( 'folder_name' );
 $config_obj->removefolder( 'folder_name' );
 ### list folders
 $config_obj->listfolders();

 ### add, retrieve, and remove notes
 $config_obj->setnote( -year => 2002, -month => 1, -day => 26, -note => 'Happy Birthday' );
 my $note_string = $config_obj->getnote( -year => 2002, -month => 1, -day => 26 );
 $config_obj->removenote( -year => 2002, -month => 1, -day => 26 );

 ### add, list, and remove contacts
 $config_obj->setcontact( -email => 'johndoe@publicdomain.com', attribute1 => 'value', attribute2 => 'value' );
 my %contacts = $config_obj->listcontacts();
 $config_obj->removecontact( -email => 'johndoe@publicdomain.com' );

=head1 CONSTRUCTOR

=over 4

=item I<-debug>

 Description: turns debugging on or off [default off].

=item I<-user_storage_dir>

 Description:  Base storage directory

=item I<-email>

 Description: email address (used for storage folder)

=back

=head1 METH0DS

=over 4

=item I<new> Constructor

 Description: 

=item I<config> Returns config as hashref

 Description: 

=item I<read_config>

 Description: Commits config to storage
 Arguments: none

=item I<write_config>

 Description: Commits config to storage
 Arguments: none

=item I<set>

 Description: 

=item I<addfolder>

 Description: Adds folder to internal folder list
 Arguments: 'folder name'

=item I<removefolder>

 Description: Removes folder to internal folder list
 Arguments: 'folder name'

=item I<setnote>

 Description: Adds note for specified year, month, and day
 Arguments: -year =>, -month =>, -day =>, -note =>

=item I<getnote>

 Description: Retrieves note for specified year, month, and day
 Arguments: -year =>, -month =>, -day =>, -note =>

=item I<removenote>

 Description: Removes note for specified year, month, and day
 Arguments: -year =>, -month =>, -day =>

=item I<setcontact>

 Description: Adds a contact 
 Arguments: -email => '', attribute_keys => values ..

=item I<listcontacts>

 Description: returns hash of contacts
 Arguments: none

=item I<removecontact>

 Description: Removes contact
 Arguments: -email => ''

=back

=head1 AUTHOR

 Jason Woodward <woodwardj@jaos.org>

=cut
