#!/usr/bin/perl
$| = 1;
################################################
#
# Version 3.1.0          November 30, 2005
#
# PerlWebmail - utility to check mail using regular 
# POP3 mail, sends via SMTP.  All done as simply 
# as possible.
#
# by Jason Woodward <woodwardj@jaos.org>
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
use strict;
use warnings;
use PerlWebmail::Config ();
use PerlWebmail::Language ();
use PerlWebmail::UserConfig ();
use Mail::Box::Manager ();
use Mail::Box::Search::Grep ();
use Mail::Transport::SMTP ();
use Mail::Message::Convert::HtmlFormatText ();
use Crypt::Blowfish ();
use Crypt::CBC ();
use Template ();
use CGI ();
use Date::Parse ();
use Text::Wrap ();
use POSIX ();
use Data::Dumper ();
use constant DIRDELIM => '/';

our $TRUN_MESSAGE_PART_LENGTH = 50;

handler() unless exists $ENV{MOD_PERL};

sub handler {
  my $request = build_request();

  unless ($request->{logged_in} && $request->{have_valid_action}) {
    finish_request($request,'login')
  } else {
    $request->{action}->($request);
  }

  close_open_resources($request);

}

sub load_config
{
  my ($r) = @_;

  if ( $r->{cgi}->server_name() && $PerlWebmail::Config::config->{ $r->{cgi}->server_name() } ) {
    return $PerlWebmail::Config::config->{ $r->{cgi}->server_name() };
  } else {
    return $PerlWebmail::Config::config->{default};
  }

}

sub load_language
{
  my ($r) = @_;


  my $lang;

  if ( $r->{user_config} && defined $r->{user_config}->{language} ) {
    $lang = $r->{user_config}->{language};
  } else {
    $lang = $r->{cgi}->http('HTTP_ACCEPT_LANGUAGE');
  }

  if ( defined $PerlWebmail::Language::language->{$lang} ) {
    return $PerlWebmail::Language::language->{$lang};
  }
  
  return $PerlWebmail::Language::language->{en};
}

sub finish_request
{
  my ($request,$tmpl) = @_;

  # this will move to a global var so in persistent we save overhead
  my $template = Template->new({
    INCLUDE_PATH => $request->{config}->{template_dir}
  });

  print $request->{cgi}->header(
    -cache_control => 'no-cache',
    -pragma => 'no-cache',
    -expires => 'now',
    -charset => $request->{config}->{charset},
  );

  eval {
    $template->process(
      $tmpl,
      {
        request                     => $request,
        config                      => $request->{config},
        user_config                 => $request->{user_config},
        language                    => $request->{language},
        gen_auth_link               => sub { gen_auth_link(@_) },
        gen_auth_hidden_form_fields => sub { gen_auth_hidden_form_fields(@_) },
        strftime                    => sub { POSIX::strftime(@_) },
      },
    ) or die "$@";
  };

  if ($@) {
    print "<h1>$@</h1>";
  }

}

sub load_parameters
{
  my ($r) = @_;

  my $params = {};
  my $local_cgi = undef;
  if ( $r->{cgi}->param('encrypted') ) {
    $local_cgi = CGI->new(
      $r->{cipher_handle}->decrypt_hex(
        $r->{cgi}->param('encrypted')
      )
    );
  } else {
    $local_cgi = $r->{cgi};
  }

  foreach my $obj ( ($r->{cgi},$local_cgi) ) {

    foreach my $p ( $obj->param ) {
      next if $p eq 'encrypted';

      my @values = $obj->param($p);

      if ( @values == 1 ) {
        $params->{$p} = $values[0];
      } else {
        $params->{$p} = [@values];
      }

    }

  }

  return $params;
}

sub build_request
{
  my $r = {};

  $r->{cgi} = CGI->new();

  $r->{config} = load_config($r);

  # use compat calling convention for Crypt::CBC
  $r->{cipher_handle} = Crypt::CBC->new({
    # too specific for certain proxied users
    #-key => $r->{cgi}->remote_host() . $r->{cgi}->user_agent(),
    # only use the first three octets
    key => join('.', (split('\.',$r->{cgi}->remote_host()))[0,1,2]) . $r->{cgi}->user_agent(),
    cipher => 'Blowfish',
    salt => 1
  });

  $r->{param} = load_parameters($r);

  $r->{actions} = load_actions($r);

  if ( $r->{param}->{action} && $r->{actions}->{$r->{param}->{action}} ) {
    $r->{have_valid_action} = 1;
    $r->{action} = $r->{actions}->{$r->{param}->{action}};
  }

  $r->{webserver} = $r->{cgi}->server_name();
  $r->{script_name} = $r->{cgi}->script_name();
  $r->{current_folder} = $r->{param}->{folder};
  ($r->{attachment}) = $r->{attachment} =~ m!([^/\\:]+)\z! if $r->{attachment};

  # find a password
  my $password = $r->{param}->{password} || $r->{param}->{encrypted_password};
  $password = $r->{cipher_handle}->decrypt_hex( $r->{param}->{encrypted_password} )
    if $r->{param}->{encrypted_password};
  $r->{username} = $r->{param}->{username};

  $r->{logged_in} = undef;

  if ( $r->{username} && $password ) {

    $r->{email} = $r->{username} . '@' . $r->{config}->{hostname};
    $r->{working_dir} = $r->{config}->{user_storage} . DIRDELIM . $r->{email};
    $r->{mgr} = Mail::Box::Manager->new( folderdir => [ $r->{working_dir} ] );

    # login the user if the parameters for u/p are present
    if ( $r->{logged_in} = validate_credentials($r,$r->{username},$password) ) {

      eval {
        $r->{user_config_obj} = PerlWebmail::UserConfig->new(
          -user_storage_dir => $r->{config}->{user_storage},
          -email => $r->{email},
        ) or die "Could not initialize PerlWebmail::UserConfig object";
        $r->{user_config} = $r->{user_config_obj}->config();
        open_current_folder($r);
      };
      if ( $@ ) {
        $r->{error} = $@;
        undef $r->{logged_in};
      }
    }
  }

  $r->{language} = load_language($r);

  return $r;
}

# default actions... this allows future plugability
sub load_actions
{
  my ($r) = @_;

  return {

    show_folder         => sub { message_actions($r) if $r->{param}->{message_action}; gen_message_list($r), show_folder($r); },
    read_message        => sub { retrieve_message($r), read_message($r) },
    get_message_as_mbox => sub { get_message_as_mbox($r) },
    create_message      => sub { create_message($r) },
    send_message        => sub { send_message($r) },
    get_attachment      => sub { get_attachment($r) },
    show_calendar       => sub { close_open_resources($r), show_calendar($r) },
    show_contacts       => sub { close_open_resources($r), show_contacts($r) },
    show_settings       => sub { close_open_resources($r), show_settings($r) },
    export_contacts     => sub { close_open_resources($r), export_contacts($r) },
  };
}

sub close_open_resources
{
  my ($r) = @_;

  $r->{mgr}->closeAllFolders() if $r->{mgr};
}

sub validate_credentials
{
  my ($r,$user,$pass) = @_;
  return unless ($r && $user && $pass);

  my $tmp_lang =
    $PerlWebmail::Language::language->{$r->{cgi}->http('HTTP_ACCEPT_LANGUAGE')}
    || $PerlWebmail::Language::language->{en};

  eval {
    $r->{folders}->{inbox} = $r->{mgr}->open(
      folder => 'pop3',
      type => 'pop3',
      username => $user,
      password => $pass,
      server_name => $r->{config}->{popserver},
      server_port => $r->{config}->{popserverport},
      authenticate => $r->{config}->{popauthmode},
      access => 'rw',
      trace => 'NONE',
      log => 'NONE',
    ) or die "Failed to validate account";
  };
  if ( $@ || !$r->{mgr}->isOpenFolder($r->{folders}->{inbox}) ) {
    $r->{error} = $tmp_lang->{ERRNO_LOGIN};
    return;
  } else {
    return 1;
  }
}

# this generates the message listing from the already open current_folder
sub gen_message_list
{
  my ($r) = @_;

  my $filter;
  if ( $r->{param}->{folder_search} ) {
    my $regex;

    eval { $regex = qr/$r->{param}->{folder_search}/i; };
    if ($@) {
      $regex = $r->{param}->{folder_search};
    }

    $filter = Mail::Box::Search::Grep->new(
      in => 'HEAD', match => $regex, decode => 0,
      delayed => 1, multiparts => 0,
    );
  }

  my $sort_routines = {
    from        => sub { lc(join(',',(map { $_->format() } $a->from()))) cmp lc(join(',',(map { $_->format() } $b->from()))) },
    fromrev     => sub { lc(join(',',(map { $_->format() } $b->from()))) cmp lc(join(',',(map { $_->format() } $a->from()))) },
    date        => sub { Date::Parse::str2time($a->head->get('Date')) <=> Date::Parse::str2time($b->head->get('Date')) },
    daterev     => sub { Date::Parse::str2time($b->head->get('Date')) <=> Date::Parse::str2time($a->head->get('Date')) },
    subject     => sub { lc($a->subject() ? $a->subject : "") cmp lc($b->subject() ? $b->subject : "") },
    subjectrev  => sub { lc($b->subject() ? $b->subject : "") cmp lc($a->subject() ? $a->subject : "") },
    size        => sub { sprintf("%.2f",( $a->size() / 1024)) <=> sprintf("%.2f",( $b->size() / 1024)) },
    sizerev     => sub { sprintf("%.2f",( $b->size() / 1024)) <=> sprintf("%.2f",( $a->size() / 1024)) },
  };

  # pick the sort routine to use
  my $sr = $sort_routines->{
    $r->{param}->{sort_by} || $r->{user_config}->{defaultsort} || $r->{config}->{defaultsort}
  };

  # start from 0, or $r->{param}->{lastdisplayed} + 1
  $r->{message_list_start} = $r->{param}->{lastdisplayed} ?
    $r->{param}->{lastdisplayed} + 1 : 0;
  # figure out a stopping point
  $r->{message_list_stop} = $r->{message_list_start};
  if ($r->{user_config}->{maxmesslistnum}) {
    $r->{message_list_stop} += $r->{user_config}->{maxmesslistnum} -1;
  } elsif ($r->{config}->{maxmesslistnum}) {
    $r->{message_list_stop} += $r->{config}->{maxmesslistnum} -1;
  }
  # running count of how many messages are in the folder or search
  $r->{message_list_count} = 0;
  # running count of how many messages are new
  $r->{message_list_new} = 0;

  # perform the search and store the messages
  my @messages;
  if ($r->{param}->{folder_search} && $filter) {
    @messages = grep {not $_->isDeleted} $filter->search($r->{folders}->{ $r->{current_folder} });
  } else { # or use all messages in the folder
    @messages = $r->{folders}->{ $r->{current_folder} }->messages('ACTIVE');
  }
  $r->{message_list_count} = scalar @messages;

  # now lets sort our messages and truncate the list to the appropriate number to display
  @{$r->{messages}} = map {
    my $m = $_;
    my $o = {};

    unless ($m->label('seen')) {
      $r->{message_list_new}++;
      $o->{new} = 1;
    }

    $o->{from} = join ',',(my @a = map { $_->format() } $m->from());
    substr($o->{from},$TRUN_MESSAGE_PART_LENGTH,length($o->{from}),'...')
      if length($o->{from}) >= $TRUN_MESSAGE_PART_LENGTH;
    $o->{from} = $r->{language}->{LISTMESSAGES_EMPTYFROM} unless $o->{from};
    $o->{from} = $r->{cgi}->escapeHTML($o->{from});

    $o->{subject} = $m->subject();
    substr($o->{subject},$TRUN_MESSAGE_PART_LENGTH,length($o->{subject}),'...')
      if length($o->{subject}) >= $TRUN_MESSAGE_PART_LENGTH;
    $o->{subject} = $r->{language}->{LISTMESSAGES_NOSUBJECT} unless $o->{subject};
    $o->{subject} = $r->{cgi}->escapeHTML($o->{subject});

    $o->{date} = $m->head->get('Date');
    $o->{size} = sprintf("%.2f",( $m->size() / 1024));
    $o->{id} = $m->messageId();

    $o;
  } grep { defined } (sort $sr @messages)[$r->{message_list_start} .. $r->{message_list_stop}];

  # this is used to define lastdisplayed later
  $r->{message_list_last_displayed} = $r->{message_list_stop};
  $r->{message_list_stop} = $r->{message_list_start} + scalar(@{$r->{messages}});
  if ($r->{message_list_stop} > $r->{message_list_count}) {
    $r->{message_list_stop} = $r->{message_list_count};
  }
  $r->{message_list_start} = 0 if (scalar( @{$r->{messages}} ) == 0);

  $r->{start} = $r->{message_list_start};
  $r->{stop} = $r->{message_list_stop};
  ++$r->{start} if ($r->{stop} != 0);

  close_open_resources($r);

}

sub show_folder
{
  my ($r) = @_;

  return finish_request($r,'login') if ( defined $r->{error} );

  $r->{intl_date} = gen_intl_date($r);

  finish_request($r,'list');
}

sub gen_auth_link
{
  my ($r,$options) = @_;

  my $encrypted_password = $r->{param}->{encrypted_password} ||
     $r->{cipher_handle}->encrypt_hex($r->{param}->{password});

  my $l = "username=$r->{username}&encrypted_password=$encrypted_password";

  foreach my $field ( keys %$options ) {
    next unless $field && $options->{$field};
    $l .= qq{&$field=$options->{$field}};
  }

  if ( $r->{config}->{encryptlinksandforms} ) {
    return 'encrypted=' . $r->{cipher_handle}->encrypt_hex($l);
  }

  return $l;

}

sub gen_auth_hidden_form_fields
{
  my ($r,$options) = @_;

  my $encrypted_password = $r->{param}->{encrypted_password} ||
     $r->{cipher_handle}->encrypt_hex($r->{param}->{password});

  if ( $r->{config}->{encryptlinksandforms} ) {

    my $l = "username=$r->{username}&encrypted_password=$encrypted_password";

    foreach my $field ( keys %$options ) {
      next unless $field and $options->{$field};
      $l .= qq/&$field=$options->{$field}/;
    }

    return qq{<input type="hidden" name="encrypted" value="} . $r->{cipher_handle}->encrypt_hex($l) . '">';

  } else {

    my $l = qq!
      <input type="hidden" name="username" value="$r->{username}">
      <input type="hidden" name="encrypted_password" value="$encrypted_password">
    !;

    foreach my $field ( keys %$options ) {
      next unless $field && $options->{$field};
      $l .= qq{<input type="hidden" name="$field" value="$options->{$field}">};
    }

    return $l;
  }

}

sub gen_intl_date
{
  my ($r) = @_;

  my $month = {
    'Jan' => $r->{language}->{W_JAN},
    'Feb' => $r->{language}->{W_FEB},
    'Mar' => $r->{language}->{W_MAR},
    'Apr' => $r->{language}->{W_APR},
    'May' => $r->{language}->{W_MAY},
    'Jun' => $r->{language}->{W_JUN},
    'Jul' => $r->{language}->{W_JUL},
    'Aug' => $r->{language}->{W_AUG},
    'Sep' => $r->{language}->{W_SEP},
    'Oct' => $r->{language}->{W_OCT},
    'Nov' => $r->{language}->{W_NOV},
    'Dec' => $r->{language}->{W_DEC}
  };

  my $day = {
    'Sun' => $r->{language}->{W_SUN},
    'Mon' => $r->{language}->{W_MON},
    'Tue' => $r->{language}->{W_TUE},
    'Wed' => $r->{language}->{W_WED},
    'Thu' => $r->{language}->{W_THU},
    'Fri' => $r->{language}->{W_FRI},
    'Sat' => $r->{language}->{W_SAT}
  };

  my $tmpdate = localtime;
  my (@t) = split(/\s+/, $tmpdate);
  my (@h) = split(/:/, $t[3]);

  return (sprintf "%s %s %s %s, %02d:%02d%s",
    $day->{$t[0]}, $month->{$t[1]}, $t[2], $t[4],
    ($h[0] <= 12 ? $h[0] : $h[0] - 12),
    $h[1], ($h[0] < 12 ? 'AM' : 'PM'));
}

sub retrieve_message
{
  my ($r) = @_;

  my $folder = $r->{folders}->{ $r->{current_folder} };

  $r->{current_message_obj} = $folder->find( $r->{param}->{message_id} );
  unless ( $r->{current_message_obj} ) {
    close_open_resources($r);
    return;
  }

  # make sure its marked as seen
  $r->{current_message_obj}->label( seen => 1 );
  $r->{current_message_obj}->parts('RECURSE');

  # store message attributes
  $r->{current_message}->{headers} =
    $r->{cgi}->escapeHTML( Text::Wrap::wrap( '','', $r->{current_message_obj}->head() ) );

  my $seqnr = $r->{current_message_obj}->seqnr();
  if ( $seqnr > 1 ) {
    if ( my $prev_msg = $folder->message( $seqnr - 1) ) {
      $r->{current_message}->{prev_message} = $prev_msg->messageId();
    }
  }
  if ( my $next_msg = $folder->message( $seqnr + 1) ) {
    $r->{current_message}->{next_message} = $next_msg->messageId();
  }
  $r->{current_message}->{seqnr} = $seqnr;
  $r->{current_message}->{id} = $r->{current_message_obj}->messageId();


  $r->{current_message}->{to} =
    join(',', map {
      $r->{cgi}->escapeHTML($_->format()) .
      gen_add_contact_link($r,$_->address(), $_->name())
    } $r->{current_message_obj}->to() ) || undef;

  $r->{current_message}->{cc} =
    join(',', map {
      $r->{cgi}->escapeHTML($_->format()) .
      gen_add_contact_link($r,$_->address(), $_->name())
    } $r->{current_message_obj}->cc() ) || undef;

  $r->{current_message}->{from} =
    join(',', map {
      $r->{cgi}->escapeHTML($_->format()) .
      gen_add_contact_link($r,$_->address(), $_->name())
    } $r->{current_message_obj}->from() ) || undef;

  $r->{current_message}->{reply_to} =
    $r->{cgi}->escapeHTML( $r->{current_message_obj}->get('reply_to') ) || undef;

  $r->{current_message}->{date} =
    $r->{current_message_obj}->head->get('Date') || undef;

  $r->{current_message}->{subject} =
    $r->{cgi}->escapeHTML( $r->{current_message_obj}->subject() ) ||
    $r->{language}->{LISTMESSAGES_NOSUBJECT};

  # now pull out the message body and any attachments
  if ( $r->{current_message_obj}->isMultipart ) {
    my $part_count = 0;

    # loop through all parts of the message
    foreach my $part ( $r->{current_message_obj}->body->parts ) {
      if ( $part->body->mimeType =~ m/(?:^text)|(?:^message)/ ) {
        $r->{current_message}->{body} .= render_email($r,$part);
      } elsif ( $part->body->mimeType =~ m/(?:signed$)|(?:multipart\/)/ ) {
        foreach my $multi_part ( $part->body->parts ) {
          if ( $multi_part->body->mimeType =~ m/(?:^text)|(?:^message)|(?:pgp-signature$)/ ) {
            $r->{current_message}->{body} .= render_email($r,$multi_part);
          }
        }
      } else {
        # this is an attachment
        $r->{current_message}->{attachments}->{$part_count} = {
          size => $part->size,
          type => $part->body->mimeType,
          name => ($part->body->disposition =~ m/filename=(.*)/) ? $1 : '',
        };
      }
      ++$part_count;
    }

  } else {
    $r->{current_message}->{body} = render_email($r,$r->{current_message_obj});
  }

  close_open_resources($r);
}

sub render_email
{
  my ($r,$email) = @_;

  my $body = $email->body();

  if ( $email->body->mimeType eq 'text/html' && !$r->{user_config}->{html_email} ) {

    $body = html_convert( $body->decoded );
    $body = linkify( $r, Text::Wrap::wrap('','',$body ) );

  } elsif ( $email->body->mimeType eq 'text/plain' ) {

    $body = linkify( $r, Text::Wrap::wrap('','',$body->decoded ) );

  # give it a shot... this helps with plain text attachments such as
  # text/c-src etc.
  } elsif ( $email->body->mimeType =~ m/text/ ) {

    $body = linkify( $r, Text::Wrap::wrap('','',$body->decoded ) );

  }

  return $body;
}

sub linkify
{
  my ($r,$content) = @_;
  $content =~ s/<\W?script.*?<\W?\/\W?script\W?>//gi;
  $content =~ s!(http|https|ftp)(://[-a-zA-Z0-9%.~:_/&amp;=\?#]+)(\w+)!<a target="_blank" href="$1$2$3">$1$2$3</a>!g;

  my $create_message_link = gen_auth_link(
    $r,
    {
      folder => $r->{current_folder},
      action => 'create_message',
    }
  );

  $content =~ s!([-a-zA-Z0-9_.]+@([-a-zA-Z0-9]+\.)+[A-Za-z]{2,4})!<a href="mailto:$1">$1</a> <a href="?to=$1&$create_message_link">\[$r->{language}->{W_EMAIL_NOW}\]</a>!xg;

  return $content;
}

sub html_convert 
{
  my ($content) = @_;
  $content =~ s/<\W?script.*?<\W?\/\W?script\W?>//gi;
  return Mail::Message::Convert::HtmlFormatText->new->format($content);
}

sub gen_add_contact_link
{
  my ($r,$email,$full_name) = @_;

  return '' unless ( $email ); # $full_name may or may not be passed in

  my $add_contact_link = gen_auth_link(
    $r,
    {
      action => 'read_message',
      message_id => $r->{current_message}->{id},
      folder => $r->{current_folder},
      add_contact => 1,
      contact_email => $email,
      contact_full_name => $full_name,
    }
  );

  return qq{&nbsp;
    <a href="?$add_contact_link">[$r->{language}->{READMESSAGE_ADDCONTACT}]</a> &nbsp; };
}

sub read_message
{
  my ($r) = @_;

  # edit/create contact here
  if ( $r->{param}->{add_contact} ) {
    $r->{user_config_obj}->setcontact(
      -email => $r->{param}->{contact_email},
      full_name => $r->{param}->{contact_full_name},
    );
  }

  return finish_request($r,'login') if ( defined $r->{error} );
  finish_request($r,'read_message');
}

sub get_message_as_mbox
{
  my ($r) = @_;

  ### get the message in question
  my $msg = $r->{folders}->{ $r->{current_folder} }->find( $r->{param}->{message_id} );
    if ( $msg ) {
    ### print a simple header, and the message head and decoded body
    print $r->{cgi}->header(
      -type => 'text/plain',
      -content_disposition => "inline; filename=$r->{param}->{message_id}",
    ),$msg->head(),$msg->decoded();
  }

}

sub create_message
{
  my ($r) = @_;
  return finish_request($r,'login') if ( defined $r->{error} );

  $r->{header} = "$r->{language}->{ERROR_SEND} $r->{send_error}<br>" if $r->{send_error};

  # if we don't already have the message from previous submit, build it
  unless ( $r->{send_error} ) {
    my $msg;

    # build forward
    if ( $r->{param}->{as_forward_from} ) {

      $r->{header} = $r->{language}->{READMESSAGE_FORWARD};

      if ( my $base_msg = $r->{folders}->{ $r->{current_folder} }->find( $r->{param}->{as_forward_from} ) ) {
        $msg = $base_msg->forward(
          To => $r->{email}, # Mail::Box requires a 'To' for some reason :-/
          Subject => $base_msg->forwardSubject( $base_msg->subject() ),
          include => $r->{user_config}->{reply_style} ? $r->{user_config}->{reply_style} : 'INLINE',
          group_reply => 1,
        );
        $msg->head->set('to',undef); # Mail::Box requires a 'To' for some reason :-/
      }

    # build reply
    } elsif ( $r->{param}->{as_reply_from} ) {

      $r->{header} = $r->{language}->{READMESSAGE_REPLY};

      if ( my $base_msg = $r->{folders}->{ $r->{current_folder} }->find( $r->{param}->{as_reply_from} ) ) {
        $msg = $base_msg->reply(
          To => $base_msg->get('reply-to') || $base_msg->sender() || $base_msg->from(),
          Subject => $base_msg->replySubject( $base_msg->subject() ),
          include => $r->{user_config}->{reply_style} ? $r->{user_config}->{reply_style} : 'INLINE',
          group_reply => 1,
        );
      }

    }

    # extract for form population
    if ( $msg ) {
      $r->{param}->{to}       = join(',',map { $_->address() } $msg->to() ) if $msg->to();
      $r->{param}->{cc}       = join(',',map { $_->address() } $msg->cc() ) if $msg->cc();
      $r->{param}->{bcc}      = join(',',map { $_->address() } $msg->bcc() ) if $msg->bcc();
      $r->{param}->{reply_to} = $msg->get('reply-to') if $msg->get('reply-to');
      $r->{param}->{subject}  = $msg->subject() if $msg->subject();


      # setup body without attachments
      if ( $msg->body() ) {
        if ( $msg->body->isMultipart() ) {
          foreach my $part ( $msg->body->parts ) {

            if ($part->body->mimeType =~ m/(?:^text)|(?:^message)/ ) {
              $r->{param}->{body} .= "> " . $part->body() . "\n";
            } elsif ( $part->body->mimeType =~ m/(?:signed$)|(?:multipart)/ ) {
              # go through the multipart message looking for textual bodies
              foreach my $signed_part ( $part->body->parts ) {
                if ($signed_part->body->mimeType =~ m/(?:^text)|(?:^message)|(?:pgp-signature$)/) {
                  my $body = $signed_part->body();
                  $body =~ s/^/\> /gm;
                  $r->{param}->{body} .= $body . "\n";
                }
              }
            }

          }# foreach my $part
        } else {
          $r->{param}->{body} = $msg->body();
        }
      }

    }

  }

  $r->{param}->{to} = join(',',@{$r->{param}->{to}})
    if ( ref $r->{param}->{to} eq 'ARRAY' );
  $r->{param}->{cc} = join(',',@{$r->{param}->{cc}})
    if ( ref $r->{param}->{cc} eq 'ARRAY' );
  $r->{param}->{bcc} = join(',',@{$r->{param}->{bcc}})
    if ( ref $r->{param}->{bcc} eq 'ARRAY' );

  # attach signature if available
  $r->{param}->{body} .= "\n\n--\n$r->{user_config}->{signature}\n"
    if $r->{user_config}->{signature};

  #$r->{param}->{body} = Mail::Message::Body->new( data => $r->{param}->{body} . "\n" )
  #  if $r->{param}->{body};
  close_open_resources($r);
  finish_request($r,'create_message');
}

sub send_message
{
  my ($r) = @_;

  # don't send empty messages
  return create_message($r)
    unless ( $r->{param}->{to} or $r->{param}->{cc} or $r->{param}->{bcc} );

  # construct the attachments and the msg
  my $attachment;
  if ( $r->{param}->{attachment} ) {
    my $attachment_data;
    my $attachment_fh = $r->{cgi}->upload('attachment');
    # slurp in the attachment
    {
      local $/ = undef;
      $attachment_data = <$attachment_fh>;
    }

    my $disposition = Mail::Message::Field->new (
      'Content-Disposition' => 'attachment',
      filename => $r->{cgi}->param('attachment'),
    );

    require IO::Scalar;
    $attachment = Mail::Message::Body::Lines->new(
      file => IO::Scalar->new(\$attachment_data),
      disposition => $disposition,
      mime_type => $r->{cgi}->uploadInfo( my $tmpfile = $r->{cgi}->param('attachment') )->{'Content-Type'},
    );
  }

  my $attach_orig = 1 if (
    ($r->{param}->{as_reply_from} || $r->{param}->{as_forward_from})
    && $r->{user_config}->{reply_style} eq 'ATTACH'
  );
  my $orig_id = $r->{param}->{as_reply_from} || $r->{param}->{as_forward_from};

  my $msg = Mail::Message->build(
    From    => $r->{user_config}->{full_name}
              ? Mail::Address->new($r->{user_config}->{full_name}, $r->{email} )
              : Mail::Address->new($r->{email},$r->{email}),
    To      => [ Mail::Address->parse($r->{param}->{to}) ],
    data    => $r->{param}->{body} . "\n",
    attach  => $attachment || undef,
    attach  => $attach_orig
              ? $r->{folders}->{ $r->{current_folder} }->find( $orig_id )
              : undef,
  );
  $msg->head->set('cc',[Mail::Address->parse($r->{param}->{cc})]) if $r->{param}->{cc};
  $msg->head->set('bcc',[Mail::Address->parse($r->{param}->{bcc})]) if $r->{param}->{bcc};
  $msg->head->set('subject',$r->{param}->{subject}) if $r->{param}->{subject};
  $msg->head->set('Reply-To',$r->{param}->{replyto}) if $r->{param}->{replyto};
  $msg->head->set('X-Mailer','PerlWebmail version 3 (http://software.jaos.org)');
  if ( $r->{param}->{as_reply_from} && $orig_id ) {
    $msg->head->set('References',$orig_id);
    $msg->head->set('In-Reply-To',$orig_id);
  }

  # attempt to send it
  eval {
    my $transport = Mail::Transport::SMTP->new(
      via => 'smtp',
      helo_domain => $r->{config}->{hostname},
      hostname => $r->{config}->{smtpserver},
      port => $r->{config}->{smtpserverport},
      interval => 3, retry => 3, timeout => 3,
      # username => $r->{param}->{username},
      # password => $r->{param}->{password},
    ) or die "Failed to create transport: [$!]";
    $transport->send($msg) or die "Send failed: [$!]";
  };
  if ($@) {
    $r->{send_error} = $@;
    return create_message($r);
  }

  if ( $r->{user_config}->{copyoutgoingto} ) {
    my $fld = $r->{folders}->{$r->{user_config}->{copyoutgoingto}} ||
      $r->{mgr}->open($r->{user_config}->{copyoutgoingto}, access => 'rw' );

    if ( $fld ) {
      $msg->label( seen => 1 );
      $r->{mgr}->appendMessage( $fld, $msg );
    }
  }

  $r->{actions}->{show_folder}->($r);
}

sub get_attachment
{
  my ($r) = @_;
  my $msg = $r->{folders}->{ $r->{current_folder} }->find( $r->{param}->{message_id} );
  return read_message($r) unless $msg;

  $msg->parts('RECURSE');
  my @parts = $msg->parts;
  my $part = $parts[$r->{param}->{part_number}];
  print $r->{cgi}->header(
    -type => $part->body->mimeType,
    -content_length => $part->size(),
    -content_disposition =>
      ( $part->body->disposition =~ m/filename=(.*)/ )
        ? "attachment; filename=$1"
        : "attachment; filename=" . $part->body->mimeType,
  ),$part->body->decoded();

}

sub show_calendar
{
  my ($r) = @_;

  $r->{calendar}->{months} = {
    1 => ['January',31],
    2 => ['Febuary',28],
    3 => ['March',31],
    4 => ['April',30],
    5 => ['May',31],
    6 => ['June',30],
    7 => ['July',31],
    8 => ['August',31],
    9 => ['September',30],
    10 => ['October',31],
    11 => ['November',30],
    12 => ['December',31],
  };

  $r->{calendar}->{month} = $r->{param}->{month} ? $r->{param}->{month} : (localtime())[4] + 1;
  $r->{calendar}->{year} = $r->{param}->{year} ? $r->{param}->{year} : (localtime())[5] + 1900;
  $r->{calendar}->{day} = $r->{param}->{day} ? $r->{param}->{day} : (localtime)[3];
  $r->{calendar}->{months}->{2}->[1] += 1 unless ($r->{calendar}->{year} % 4); # compensate for leap year
  # sanity check, month day can't be greater than last day of month
  if ($r->{calendar}->{day} > $r->{calendar}->{months}->{$r->{calendar}->{month}}->[1]) {
    $r->{calendar}->{day} = $r->{calendar}->{months}->{$r->{calendar}->{month}}->[1];
  }
  # use POSIX::strftime() to get day of week
  $r->{calendar}->{day_of_week} = POSIX::strftime("%A",0,0,0,$r->{calendar}->{day},($r->{calendar}->{month} - 1),($r->{calendar}->{year}-1900));

  # set the note
  if ( $r->{param}->{note} && $r->{param}->{day} && $r->{param}->{month} && $r->{param}->{year} ) {
    # commit the note
    if ( $r->{param}->{commit} ) {
      $r->{user_config_obj}->setnote(
        -year => $r->{param}->{year},
        -month => $r->{param}->{month},
        -day => $r->{param}->{day},
        -note => $r->{param}->{note},
      );
    # or delete the note
    } elsif ( $r->{param}->{delete} ) {
      $r->{user_config_obj}->removenote(
        -year => $r->{param}->{year},
        -month => $r->{param}->{month},
        -day => $r->{param}->{day},
      );
    }
  }

  $r->{note_string} = $r->{user_config_obj}->getnote(
    -day => $r->{calendar}->{day},
    -month => $r->{calendar}->{month},
    -year => $r->{calendar}->{year},
  ) || '';

  finish_request($r,'calendar');
}

sub show_contacts
{
  my ($r) = @_;

  # delete contacts here
  if ( $r->{param}->{delete_contact_submit} && $r->{param}->{contacts_delete} ) {
    my $contacts = $r->{param}->{contacts_delete};
    if ( ref $contacts eq 'ARRAY' ) {
      foreach my $email ( @$contacts ) {
        next unless $email;
        $r->{user_config_obj}->removecontact( -email => $email );
      }
    } else {
      $r->{user_config_obj}->removecontact( -email => $contacts );
    }
  }# end delete contacts

  # edit/create contact here
  if ( $r->{param}->{create_contact_submit} && $r->{param}->{contact_email} ) {
    $r->{user_config_obj}->setcontact(
      -email => $r->{param}->{contact_email},
      full_name => $r->{param}->{contact_full_name},
    );
  }

  my %contacts = $r->{user_config_obj}->listcontacts();
  foreach my $contact ( sort { ($contacts{$a}->{full_name} || '') . $a cmp ($contacts{$b}->{full_name} || '') . $b } keys %contacts ) {
    next unless ($contact && $contacts{$contact});
    $r->{contact_select_options} .= qq{<option value="$contact">};
    $r->{contact_select_options} .= $r->{cgi}->escapeHTML("$contacts{$contact}->{full_name} <$contact>");
    $r->{contact_select_options} .= qq{</option>};
  }

  if ( $r->{param}->{contacts_to_edit} ) {
    $r->{edit_contact}->{email} = $r->{param}->{contacts_to_edit};
    $r->{edit_contact}->{full_name} = $contacts{ $r->{param}->{contacts_to_edit} }->{full_name};
  }

  finish_request($r,'contacts');
}

sub show_settings
{
  my ($r) = @_;
  return finish_request($r,'login') if ( defined $r->{error} );

  # if $r->{param}->{update_user_config} is set
  # we update the users configuration settings here
  if ( $r->{param}->{update_user_config} ) {
    $r->{user_config}->{maxmesslistnum} = $r->{param}->{maxmesslistnum} if $r->{param}->{maxmesslistnum};
    $r->{user_config}->{defaultsort} = $r->{param}->{defaultsort} if $r->{param}->{defaultsort};
    $r->{user_config}->{reply_style} = $r->{param}->{reply_style} if $r->{param}->{reply_style};
    $r->{user_config}->{signature} = $r->{param}->{signature} if defined $r->{param}->{signature};
    $r->{user_config}->{full_name} = $r->{param}->{full_name} if defined $r->{param}->{full_name};
    $r->{user_config}->{copyoutgoingto} = $r->{param}->{copyoutgoingto} if defined $r->{param}->{copyoutgoingto};
    $r->{user_config}->{ondeletemoveto} = $r->{param}->{ondeletemoveto} if defined $r->{param}->{ondeletemoveto};
    $r->{user_config}->{language} = $r->{param}->{language} if defined $r->{param}->{language};
    $r->{user_config}->{html_email} = $r->{param}->{html_email} if defined $r->{param}->{html_email};

    # this will auto commit the user config
    $r->{user_config_obj}->config( $r->{user_config} );
  }

  # create_folder
  if ( $r->{param}->{create_folder} && $r->{param}->{create_folder} !~ m/^[^\w]/ ) {
    my $folder = $r->{mgr}->open( $r->{param}->{create_folder}, create => 1, access => 'rw',) or warn $!;

    if ( $folder ) {
      $r->{user_config_obj}->addfolder( $r->{param}->{create_folder} );
      $folder->close();
    }

  }

  # delete_folder
  if ( $r->{param}->{delete_folder} ) {
    my $f = $r->{mgr}->open( $r->{param}->{delete_folder}, access => 'rw',) or warn $!;

    if ( $f ) {
      $f->delete();
      $f->close();
      $r->{user_config_obj}->removefolder( $r->{param}->{delete_folder} );
    }
  }

  # empty folder
  if ( $r->{param}->{empty_folder} && $r->{param}->{empty_folder} !~ m/^[^\w]/ ) {
    my $f = $r->{mgr}->open( $r->{param}->{empty_folder}, access => 'rw',) or warn $!;

    if ( $f ) {
      $f->delete();
      $f->close();
      if ( $r->{param}->{empty_folder} ne 'inbox' ) {
        my $nf = $r->{mgr}->open( $r->{param}->{empty_folder}, create => 1, access => 'rw' );
        $nf->close();
      }
    }
  }

  # export folder
  if ( $r->{param}->{export_folder} ) {
    my $file = $r->{working_dir} . DIRDELIM . $r->{param}->{export_folder};
    eval {
      if ( -f $file ) {
        open(my $fh,"<$file") or die $!;

        print $r->{cgi}->header(
          -type => 'text/plain',
          -Content_length => -s $file,
          -content_disposition => "attachment; filename=$r->{param}->{export_folder}.mbox",
        );

        while ( defined(my $line = <$fh>) ) {
          print $line;
        }

        close($fh);
      }
    };
    unless ($@) { 
      return;
    }
  }

  # sensible defaults
  $r->{user_config}->{maxmesslistnum} ||= $r->{config}->{maxmesslistnum};
  $r->{user_config}->{defaultsort} ||= $r->{config}->{defaultsort};
  $r->{user_config}->{reply_style} ||= 'INLINE';
  $r->{user_config}->{language} ||= $r->{cgi}->http("HTTP_ACCEPT_LANGUAGE") || 'en';
  $r->{user_config}->{html_email} ||= '0';

  $r->{languages} = [sort keys %{ $PerlWebmail::Language::language }];

  finish_request($r,'settings');
}

sub export_contacts
{
  my ($r) = @_;

  print $r->{cgi}->header(
    -type => 'text/plain',
    -content_disposition => "attachment; filename=$r->{email}.vcf",
  );

  my $contacts = $r->{param}->{contacts_export};
  my @contacts_to_export;
  if ( ref $contacts eq 'ARRAY' ) {
    @contacts_to_export = @{$contacts};
  } else {
    push @contacts_to_export, $contacts;
  }

  my %contacts_in_storage = $r->{user_config_obj}->listcontacts();

  foreach my $c ( @contacts_to_export ) {
    next unless $c && $contacts_in_storage{$c};

    ### print a vCard for each selected contact
    print "BEGIN:VCARD\n";
    print "VERSION:2.1\n";
    print "FN:$contacts_in_storage{$c}->{full_name}\n";
    print "EMAIL;INTERNET:$c\n";
    print "END:VCARD\n";
    print "\n";
  }

}

sub message_actions
{
  my ($r) = @_;
  my $p = $r->{param};
  my $current_folder = $r->{folders}->{ $r->{current_folder} };
  return unless $current_folder;

  my @message_ids = ();
  if ( $p->{message_id} ) {
    @message_ids = (ref $p->{message_id} eq 'ARRAY')
      ? @{$p->{message_id}}
      : ($p->{message_id});
  }

  my ($delete_to,$del_fld) = (undef,undef);
  if ( $delete_to = $r->{user_config}->{ondeletemoveto} ) {
    $del_fld = $r->{folders}->{$delete_to} || undef;
  }
  my ($dest_name,$dest_fld) = (undef,undef);
  if ( $dest_name = $p->{message_destination} ) {
    $dest_fld = $r->{folders}->{$dest_name} || undef;
  }

  if ( $p->{message_action} eq 'delete' ) {
    $del_fld ||= $r->{mgr}->open($delete_to, access => 'rw') if $delete_to;

    foreach my $id ( @message_ids ) {
      if ( my $msg = $current_folder->find($id) ) {
        if ( $del_fld ) {
          $msg->label( seen => 1 );
          $r->{mgr}->moveMessage( $del_fld, $msg );
        } else {
          $msg->delete();
        }
      }
    }

  } elsif ( $p->{message_action} eq 'delete_all' ) {
    $del_fld ||= $r->{mgr}->open($delete_to, access => 'rw') if $delete_to;

    foreach my $msg ( $current_folder->messages() ) {
      if ( $del_fld ) {
        $msg->label( seen => 1 );
        $r->{mgr}->moveMessage( $del_fld, $msg );
      } else {
        $msg->delete();
      }
    }
    if ( $r->{current_folder} ne 'inbox' ) {
      $current_folder->close();
      $r->{folders}->{ $r->{current_folder} } =
        $r->{mgr}->open( $r->{current_folder}, create => 1, access => 'rw' );
    }

  } elsif ( $p->{message_action} eq 'markasread' ) {

    foreach my $id ( @message_ids ) {
      if ( my $msg = $current_folder->find($id) ) {
        $msg->label( seen => 1 );
        # setting the label doesn't seem to do it, so we touch the message to bypass laziness
        if( $msg->isMultipart ){
          scalar($msg->body->parts);
        }else{
          my $tmp_type = $msg->body->mimeType;
        }
      }
    }

  } elsif ( $p->{message_destination} && $dest_name ) {
    $dest_fld ||= $r->{mgr}->open($dest_name, access => 'rw');

    if ( $dest_fld ) {

      foreach my $id ( @message_ids ) {
        if ( my $msg = $current_folder->find($id) ) {
          $msg->label( seen => 1 );
          if ( $p->{message_action} eq 'copy' ) {
            $r->{mgr}->copyMessage( $dest_fld, $msg );
          } elsif ( $p->{message_action} eq 'move' ) {
            $r->{mgr}->moveMessage( $dest_fld, $msg );
          }
        }
      }

    }

  }

  # don't close these folders, they'll be closed at the end of the request
  #$del_fld->close() if $del_fld;
  #$dest_fld->close() if $dest_fld;

}

sub open_current_folder
{
  my ($r) = @_;

  die "No working folder" unless $r->{current_folder};

  return if $r->{folders}->{$r->{current_folder}} &&
    $r->{mgr}->isOpenFolder($r->{folders}->{$r->{current_folder}}); 

  my $fld = $r->{mgr}->open( $r->{current_folder}, access => 'rw' ) or die $!;

  $r->{folders}->{$r->{current_folder}} = $fld;

}

1;
